import abjad


def test_mutate__are_contiguous_logical_voice_01():
    """
    Components that start at the same moment are bad.
    Even if components are all part of the same logical voice.
    """

    voice = abjad.Voice(
        r"""
        {
            c'8
            d'8
        }
        \new Voice {
            e'8
            f'8
        }
        {
            g'8
            a'8
        }
        """
    )

    voices = [voice, voice[0]]
    assert not abjad.mutate._are_contiguous_logical_voice(voices)
    voices = list(voice[0:1]) + list(voice[0])
    assert not abjad.mutate._are_contiguous_logical_voice(voices)
    voices = list(voice[-1:]) + list(voice[-1])
    assert not abjad.mutate._are_contiguous_logical_voice(voices)


def test_mutate__are_contiguous_logical_voice_02():
    """
    Is true for strictly contiguous leaves in same staff.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    assert abjad.mutate._are_contiguous_logical_voice(staff[:])


def test_mutate__are_contiguous_logical_voice_03():
    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]
    assert abjad.mutate._are_contiguous_logical_voice(notes)


def test_mutate__are_contiguous_logical_voice_04():
    """
    Is false for time-reordered leaves in staff.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    leaves = list(staff[2:]) + list(staff[:2])
    assert not abjad.mutate._are_contiguous_logical_voice(leaves)


def test_mutate__are_contiguous_logical_voice_05():
    """
    Is true for unincorporated component.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    assert abjad.mutate._are_contiguous_logical_voice([staff])


def test_mutate__are_contiguous_logical_voice_06():
    """
    Is true for empty input.
    """

    assert abjad.mutate._are_contiguous_logical_voice([])


def test_mutate__are_contiguous_logical_voice_07():
    """
    False when components belonging to same logical voice are ommitted.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8 g'8 a'8")
    abjad.beam(voice[:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            g'8
            a'8
            ]
        }
        """
    )

    leaves = list(voice[:2]) + list(voice[-2:])
    assert not abjad.mutate._are_contiguous_logical_voice(leaves)


def test_mutate__are_contiguous_logical_voice_08():
    """
    False when components belonging to same logical voice are ommitted.
    """

    voice = abjad.Voice(
        r"""
        {
            c'8
            [
            d'8
        }
        {
            e'8
            f'8
        }
        {
            g'8
            a'8
            ]
        }
        """
    )

    components = list(voice[:1]) + list(voice[-1:])
    assert not abjad.mutate._are_contiguous_logical_voice(components)


def test_mutate__are_contiguous_same_parent_01():
    """
    Is true for strictly contiguous leaves in voice.
    Is false for other time orderings of leaves in voice.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")

    assert abjad.mutate._are_contiguous_same_parent(voice[:])

    other = list(reversed(voice[:]))
    assert not abjad.mutate._are_contiguous_same_parent(other)

    components = []
    components.extend(voice[2:])
    components.extend(voice[:2])
    assert not abjad.mutate._are_contiguous_same_parent(components)

    components = []
    components.extend(voice[3:4])
    components.extend(voice[:1])
    assert not abjad.mutate._are_contiguous_same_parent(components)
    components = [voice]
    components.extend(voice[:])
    assert not abjad.mutate._are_contiguous_same_parent(components)


def test_mutate__are_contiguous_same_parent_02():
    """
    Is true for unincorporated components when orphans allowed.
    Is false for unincorporated components when orphans not allowed.
    """

    voice = abjad.Voice(
        r"""
        {
            c'8
            d'8
        }
        {
            e'8
            f'8
        }
        """
    )

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
            }
            {
                e'8
                f'8
            }
        }
        """
    )

    assert abjad.mutate._are_contiguous_same_parent([voice])

    assert abjad.mutate._are_contiguous_same_parent(voice[:])

    assert abjad.mutate._are_contiguous_same_parent(voice[0][:])
    assert abjad.mutate._are_contiguous_same_parent(voice[1][:])

    leaves = abjad.select.leaves(voice)
    assert not abjad.mutate._are_contiguous_same_parent(leaves)


def test_mutate__are_contiguous_same_parent_03():
    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]

    assert abjad.mutate._are_contiguous_same_parent(notes)


def test_mutate__are_contiguous_same_parent_04():
    """
    Empty selection returns true.
    """

    assert abjad.mutate._are_contiguous_same_parent([])


def test_mutate__fuse_leaves_by_immediate_parent_01():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    staff = abjad.Staff([abjad.Container("c'8 c'8"), abjad.Container("c'8 c'8")])
    leaves = abjad.select.leaves(staff)
    abjad.tie(leaves)

    logical_tie = abjad.get.logical_tie(leaves[1])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                c'4
                ~
            }
            {
                c'4
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert len(result) == 2
    assert abjad.wf.wellformed(staff)


def test_mutate__fuse_leaves_by_immediate_parent_02():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    staff = abjad.Staff("c'8 c'8 c'8 c'8")
    abjad.tie(staff[:])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            ~
            c'8
            ~
            c'8
            ~
            c'8
        }
        """
    ), print(abjad.lilypond(staff))

    logical_tie = abjad.get.logical_tie(staff[1])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'2
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_03():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    note = abjad.Note("c'4")
    logical_tie = abjad.get.logical_tie(note)
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)
    assert len(result) == 1
    assert abjad.wf.wellformed(note)


def test_mutate__fuse_leaves_by_immediate_parent_04():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    voice = abjad.Voice(r"c'16 ~ c'16 ~ c'16 ~ c'16 ~ c'16 r16 r16 r16 r4 r4")
    logical_tie = abjad.get.logical_tie(voice[0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            ~
            c'16
            r16
            r16
            r16
            r4
            r4
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_05():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    voice = abjad.Voice(r"c'16 ~ c'16 ~ c'16 ~ c'16 ~ c'16 r16 r16 r16 r4 r4")
    abjad.attach(abjad.BeforeGraceContainer("b'16"), voice[0])
    logical_tie = abjad.get.logical_tie(voice[0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \grace {
                b'16
            }
            c'4
            ~
            c'16
            r16
            r16
            r16
            r4
            r4
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_06():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    voice = abjad.Voice(r"c'16 ~ c'16 ~ c'16 ~ c'16 ~ c'16 r16 r16 r16 r4 r4")
    logical_tie = abjad.get.logical_tie(voice[0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            ~
            c'16
            r16
            r16
            r16
            r4
            r4
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_07():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    voice = abjad.Voice(r"\tuplet 13/8 { \time 4/4 c'8 ~ c'8 ~ c'16 ~ c'32 r16 } r4 r2")
    logical_tie = abjad.get.logical_tie(voice[0][0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                \tuplet 13/8
                {
                    \time 4/4
                    c'4
                    ~
                    c'16.
                    r16
                }
                r4
                r2
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_08():
    """
    Fuse leaves in logical tie with same immediate parent.
    """

    staff = abjad.Staff(r"c'16 ~ c'16 ~ c'16 ~ c'16 ~ c'16 r16 r16 r16 r4 r4")
    abjad.Score([staff], name="Score")
    indicators = (
        abjad.BeforeGraceContainer("b'16"),
        abjad.Clef("alto"),
        abjad.TimeSignature((3, 4)),
        abjad.Articulation("staccato"),
        abjad.Articulation("accent"),
    )
    for indicator in indicators:
        abjad.attach(indicator, staff[0])
    logical_tie = abjad.get.logical_tie(staff[0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \grace {
                b'16
            }
            \clef "alto"
            \time 3/4
            c'4
            - \accent
            - \staccato
            ~
            c'16
            r16
            r16
            r16
            r4
            r4
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_09():
    """
    Fuse leaves in logical tie with same immediate parent, with an indicator at
    the end of the logical tie after fusing.
    """

    voice = abjad.Voice(r"d'8 c'8 ~ c'32 r16 r16 r16 r4")
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")

    indicators = (abjad.TimeSignature((3, 4)), abjad.StartBeam())
    for indicator in indicators:
        abjad.attach(indicator, voice[0])

    indicators = (
        abjad.BeforeGraceContainer("b'16"),
        abjad.Articulation("staccato"),
        abjad.Articulation("accent"),
        abjad.StopBeam(),
    )
    for indicator in indicators:
        abjad.attach(indicator, voice[1])

    logical_tie = abjad.get.logical_tie(voice[1])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                \time 3/4
                d'8
                [
                \grace {
                    b'16
                }
                c'8
                - \accent
                - \staccato
                ~
                c'32
                ]
                r16
                r16
                r16
                r4
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 1


def test_mutate__fuse_leaves_by_immediate_parent_10():
    """
    Fuse leaves in logical tie with same immediate parent, with an indicator at
    every leaf of the logical tie after fusing.
    """

    staff = abjad.Staff(r"c'16 ~ c'16 ~ c'16 ~ c'16 ~ c'16 r16 r16 r16 r4 r4")
    abjad.Score([staff], name="Score")
    indicators = (
        abjad.BeforeGraceContainer("b'16"),
        abjad.Clef("alto"),
        abjad.TimeSignature((3, 4)),
        abjad.Articulation("staccato"),
        abjad.Articulation("accent"),
        abjad.StemTremolo(16),
    )
    for indicator in indicators:
        abjad.attach(indicator, staff[0])
    logical_tie = abjad.get.logical_tie(staff[0])
    result = abjad.mutate._fuse_leaves_by_immediate_parent(logical_tie)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \grace {
                b'16
            }
            \clef "alto"
            \time 3/4
            c'4
            :16
            - \accent
            - \staccato
            ~
            c'16
            :16
            r16
            r16
            r16
            r4
            r4
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 1


def test_mutate__immediately_precedes_01():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")

    assert abjad.mutate._immediately_precedes(voice[0], voice[1])
    assert abjad.mutate._immediately_precedes(voice[1], voice[2])
    assert abjad.mutate._immediately_precedes(voice[2], voice[3])


def test_mutate__immediately_precedes_02():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")

    assert abjad.mutate._immediately_precedes(staff[0], staff[1])
    assert abjad.mutate._immediately_precedes(staff[1], staff[2])
    assert abjad.mutate._immediately_precedes(staff[2], staff[3])


def test_mutate__immediately_precedes_03():
    container = abjad.Container("c'8 d'8 e'8 f'8")

    assert abjad.mutate._immediately_precedes(container[0], container[1])
    assert abjad.mutate._immediately_precedes(container[1], container[2])
    assert abjad.mutate._immediately_precedes(container[2], container[3])


def test_mutate__immediately_precedes_04():
    tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")

    assert abjad.mutate._immediately_precedes(tuplet[0], tuplet[1])
    assert abjad.mutate._immediately_precedes(tuplet[1], tuplet[2])


def test_mutate__immediately_precedes_05():
    voice = abjad.Voice("{ c'8 d'8 e'8 f'8 } { g'8 a'8 b'8 c''8 }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(voice[0], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0], voice[1][0])
    assert abjad.mutate._immediately_precedes(voice[0][-1], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0][-1], voice[1][0])


def test_mutate__immediately_precedes_06():
    voice = abjad.Voice(r"\tuplet 3/2 { c'8 d'8 e'8 } \tuplet 3/2 { f'8 e'8 d'8 }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                c'8
                d'8
                e'8
            }
            \tuplet 3/2
            {
                f'8
                e'8
                d'8
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(voice[0], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0], voice[1][0])
    assert abjad.mutate._immediately_precedes(voice[0][-1], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0][-1], voice[1][0])


def test_mutate__immediately_precedes_07():
    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("g'8 a'8 b'8 c''8")]
    )

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
            }
            \new Voice
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(staff[0], staff[1])
    assert abjad.mutate._immediately_precedes(staff[0], staff[1][0])
    assert abjad.mutate._immediately_precedes(staff[0][-1], staff[1])
    assert abjad.mutate._immediately_precedes(staff[0][-1], staff[1][0])


def test_mutate__immediately_precedes_08():
    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("g'8 a'8 b'8 c''8")]
    )
    staff[0].name = "foo"
    staff[1].name = "foo"

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "foo"
            {
                c'8
                d'8
                e'8
                f'8
            }
            \context Voice = "foo"
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(staff[0], staff[1])
    assert abjad.mutate._immediately_precedes(staff[0], staff[1][0])
    assert abjad.mutate._immediately_precedes(staff[0][-1], staff[1])


def test_mutate__immediately_precedes_09():
    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("g'8 a'8 b'8 c''8")]
    )
    staff[0].name = "foo"
    staff[1].name = "bar"

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "foo"
            {
                c'8
                d'8
                e'8
                f'8
            }
            \context Voice = "bar"
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(staff[0], staff[1])
    assert abjad.mutate._immediately_precedes(staff[0], staff[1][0])
    assert abjad.mutate._immediately_precedes(staff[0][-1], staff[1])


def test_mutate__immediately_precedes_10():
    staff_1 = abjad.Staff([abjad.Voice("c'8 d'8 e'8 f'8")])
    staff_2 = abjad.Staff([abjad.Voice("g'8 a'8 b'8 c''8")])
    container = abjad.Container([staff_1, staff_2])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            {
                \new Voice
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }
            }
            \new Staff
            {
                \new Voice
                {
                    g'8
                    a'8
                    b'8
                    c''8
                }
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(staff_1, staff_2)
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[0][0])

    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[0][0])

    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[0][0])


def test_mutate__immediately_precedes_11():
    upper_voice_1 = abjad.Voice("c''8 d''8 e''8 f''8")
    upper_voice_2 = abjad.Voice("g''8 a''8 b''8 c''8")
    lower_voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    lower_voice_2 = abjad.Voice("g'8 a'8 b'8 c''8")
    staff_1 = abjad.Staff([upper_voice_1, lower_voice_1])
    staff_2 = abjad.Staff([upper_voice_2, lower_voice_2])
    staff_1.simultaneous = True
    staff_2.simultaneous = True
    container = abjad.Container([staff_1, staff_2])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            <<
                \new Voice
                {
                    c''8
                    d''8
                    e''8
                    f''8
                }
                \new Voice
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>
            \new Staff
            <<
                \new Voice
                {
                    g''8
                    a''8
                    b''8
                    c''8
                }
                \new Voice
                {
                    g'8
                    a'8
                    b'8
                    c''8
                }
            >>
        }
        """
    )

    assert abjad.mutate._immediately_precedes(staff_1, staff_2)
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[0][0])
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[1])
    assert abjad.mutate._immediately_precedes(staff_1, staff_2[1][0])

    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[0][0])
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[1])
    assert abjad.mutate._immediately_precedes(staff_1[0], staff_2[1][0])

    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[0][0])
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[1])
    assert abjad.mutate._immediately_precedes(staff_1[0][-1], staff_2[1][0])

    assert abjad.mutate._immediately_precedes(staff_1[1], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[1], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[1], staff_2[0][0])
    assert abjad.mutate._immediately_precedes(staff_1[1], staff_2[1])
    assert abjad.mutate._immediately_precedes(staff_1[1], staff_2[1][0])

    assert abjad.mutate._immediately_precedes(staff_1[1][-1], staff_2)
    assert abjad.mutate._immediately_precedes(staff_1[1][-1], staff_2[0])
    assert abjad.mutate._immediately_precedes(staff_1[1][-1], staff_2[0][0])
    assert abjad.mutate._immediately_precedes(staff_1[1][-1], staff_2[1])
    assert abjad.mutate._immediately_precedes(staff_1[1][-1], staff_2[1][0])


def test_mutate__immediately_precedes_12():
    voice = abjad.Voice("{ { c'8 d'8 e'8 f'8 } } { { g'8 a'8 b'8 c''8 } }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }
            }
            {
                {
                    g'8
                    a'8
                    b'8
                    c''8
                }
            }
        }
        """
    )

    assert abjad.mutate._immediately_precedes(voice[0], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0], voice[1][0])
    assert abjad.mutate._immediately_precedes(voice[0], voice[1][0][0])

    assert abjad.mutate._immediately_precedes(voice[0][0], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0][0], voice[1][0])
    assert abjad.mutate._immediately_precedes(voice[0][0], voice[1][0][0])

    assert abjad.mutate._immediately_precedes(voice[0][0][-1], voice[1])
    assert abjad.mutate._immediately_precedes(voice[0][0][-1], voice[1][0])
    assert abjad.mutate._immediately_precedes(voice[0][0][-1], voice[1][0][0])



import pytest

import abjad


def test_score_01():
    """
    Slots constrain component attributes.
    """

    for component in (
        abjad.MultimeasureRest((1, 4)),
        abjad.Chord("<ef' cs' f''>4"),
        abjad.Container(),
        abjad.Context(),
        abjad.Note("c'4"),
        abjad.NoteHead("cs''"),
        abjad.Rest((1, 4)),
        abjad.Score(),
        abjad.Skip((1, 4)),
        abjad.Staff(),
        abjad.Tuplet((2, 3), "c'8 d'8 e'8"),
        abjad.Voice(),
    ):
        with pytest.raises(AttributeError):
            component.foo = "bar"



import textwrap

import roman

from . import enums as _enums


def capitalize_start(string: str) -> str:
    """
    Capitalizes start of string.

    ..  container:: example

        Capitalizes only ``string[0]``; leaves noninitial characters unchanged:

        >>> abjad.string.capitalize_start("violin I")
        'Violin I'

        Built-in ``str.capitalize()`` forces noninitial characters to lowercase:

        >>> "violin I".capitalize()
        'Violin i'

    """
    if not string:
        return string
    return string[0].upper() + string[1:]


def delimit_words(string: str, separate_caps: bool = False) -> list[str]:
    """
    Delimits words in string.

    ..  container:: example

        >>> abjad.string.delimit_words('scale degrees 4 and 5.')
        ['scale', 'degrees', '4', 'and', '5']

        >>> abjad.string.delimit_words('scale degrees 4and5.')
        ['scale', 'degrees', '4', 'and', '5']

        >>> abjad.string.delimit_words('scaleDegrees4and5.')
        ['scale', 'Degrees', '4', 'and', '5']

        >>> abjad.string.delimit_words('ScaleDegrees4and 5.')
        ['Scale', 'Degrees', '4', 'and', '5']

        >>> abjad.string.delimit_words('scale-degrees-4-and-5.')
        ['scale', 'degrees', '4', 'and', '5']

        >>> abjad.string.delimit_words('SCALE_DEGREES_4_AND_5.')
        ['SCALE', 'DEGREES', '4', 'AND', '5']

    ..  container:: example

        >>> abjad.string.delimit_words('one < two')
        ['one', '<', 'two']

        >>> abjad.string.delimit_words('one! two!')
        ['one', '!', 'two', '!']

    ..  container:: example

        Separates capital letters when keyword is true:

        >>> abjad.string.delimit_words('MRM')
        ['MRM']

        >>> abjad.string.delimit_words("MRM", separate_caps=True)
        ['M', 'R', 'M']

        >>> abjad.string.delimit_words('MRhM')
        ['MRh', 'M']

        >>> abjad.string.delimit_words("MRhM", separate_caps=True)
        ['M', 'Rh', 'M']

    """
    wordlike_characters = ("<", ">", "!")
    words = []
    current_word = ""
    for character in string:
        if (
            not character.isalpha()
            and not character.isdigit()
            and character not in wordlike_characters
        ):
            if current_word:
                words.append(current_word)
                current_word = ""
        elif not current_word:
            current_word = current_word + character
        elif character.isupper():
            if current_word[-1].isupper() and not separate_caps:
                current_word = current_word + character
            else:
                words.append(current_word)
                current_word = character
        elif character.islower():
            if current_word[-1].isalpha():
                current_word = current_word + character
            else:
                words.append(current_word)
                current_word = character
        elif character.isdigit():
            if current_word[-1].isdigit():
                current_word = current_word + character
            else:
                words.append(current_word)
                current_word = character
        elif character in wordlike_characters:
            if current_word[-1] in wordlike_characters:
                current_word = current_word + character
            else:
                words.append(current_word)
                current_word = character
    if current_word:
        words.append(current_word)
    return words


def from_roman(string: str) -> int:
    """
    Changes string from Roman numeral to digits.

    ..  container:: example

        >>> abjad.string.from_roman('IX')
        9
        >>> abjad.string.from_roman('ix')
        9

        Raises Roman numeral error when string is not Roman numeral:

        >>> abjad.string.from_roman('Allegro')
        Traceback (most recent call last):
            ...
        roman.InvalidRomanNumeralError...

    """
    if is_roman(string):
        return roman.fromRoman(string.upper())
    number = roman.fromRoman(string)
    return number


def is_lilypond_identifier(string: str) -> bool:
    """
    Is true when string starts with a letter and either 1. contains only letters and
    underscores thereafter, or 2. contains only letters, numbers and dots thereafter.

    ..  container:: example

        >>> abjad.string.is_lilypond_identifier("ViolinOne")
        True

        >>> abjad.string.is_lilypond_identifier("Violin_One")
        True

        >>> abjad.string.is_lilypond_identifier("Violin One")
        False

    ..  container:: example

        >>> abjad.string.is_lilypond_identifier("ViolinI")
        True

        >>> abjad.string.is_lilypond_identifier("Violin_I")
        True

        >>> abjad.string.is_lilypond_identifier("Violin I")
        False

    ..  container:: example

        >>> abjad.string.is_lilypond_identifier("Violin1")
        False

        >>> abjad.string.is_lilypond_identifier("Violin_1")
        False

        >>> abjad.string.is_lilypond_identifier("Violin 1")
        False

    ..  container:: example

        >>> abjad.string.is_lilypond_identifier("Violin.1")
        True

        >>> abjad.string.is_lilypond_identifier("Violin.1.1")
        False

        >>> abjad.string.is_lilypond_identifier("Violin.1_1")
        False

        >>> abjad.string.is_lilypond_identifier("Violin_1.1")
        False

    ..  container:: example

        >>> abjad.string.is_lilypond_identifier("Violin.1.MusicVoice.1")
        True

        >>> abjad.string.is_lilypond_identifier("Violin.1.1.MusicVoice")
        False

        >>> abjad.string.is_lilypond_identifier("Violin.1_Music_Voice")
        False

        >>> abjad.string.is_lilypond_identifier("Violin.Music_Voice_1")
        False

        >>> abjad.string.is_lilypond_identifier("Violin.1_Music_Voice")
        False

    """
    if not string:
        return False
    if string[0] in "_.":
        return False
    if string[0].isdigit():
        return False
    for character in string:
        if not (character.isalnum() or character == "_" or character == "."):
            return False
    words = string.split(".")
    if words[0].isdigit():
        return False
    previous_word_ends_in_digit = False
    for word in words:
        if previous_word_ends_in_digit and word[0].isdigit():
            return False
        if any(_.isdigit() for _ in word):
            if any(not _.isdigit() for _ in word):
                return False
        if word[-1].isdigit():
            previous_word_ends_in_digit = True
        else:
            previous_word_ends_in_digit = False
    return True


def is_roman(string: str) -> bool:
    """
    Is true when string is Roman numeral.

    ..  container:: example

        >>> abjad.string.is_roman('I')
        True

        >>> abjad.string.is_roman('II')
        True

        >>> abjad.string.is_roman('X')
        True

        >>> abjad.string.is_roman('XI')
        True

        >>> abjad.string.is_roman('C')
        True

        >>> abjad.string.is_roman('CI')
        True

        >>> abjad.string.is_roman('i')
        True

        >>> abjad.string.is_roman('F')
        False

    """
    try:
        roman.fromRoman(string.upper())
        return True
    except roman.InvalidRomanNumeralError:
        return False


def is_shout_case(string: str) -> bool:
    """
    Is true when string and is shoutcase.

    ..  container:: example

        >>> abjad.string.is_shout_case('FOO_BAR')
        True

        >>> abjad.string.is_shout_case('FooBar')
        False

    """
    return string == to_shout_case(string)


def normalize(argument: str, indent: int | str | None = None) -> str:
    """
    Normalizes string.

    ..  container:: example

        >>> string = r'''
        ...     foo
        ...         bar
        ... '''
        >>> print(string)
        <BLANKLINE>
            foo
                bar
        <BLANKLINE>

        >>> print(abjad.string.normalize(string))
        foo
            bar

        >>> print(abjad.string.normalize(string, indent=4))
            foo
                bar

        >>> print(abjad.string.normalize(string, indent='* '))
        * foo
        *     bar

    """
    string = argument.replace("\t", "    ")
    lines = string.split("\n")
    while lines and (not lines[0] or lines[0].isspace()):
        lines.pop(0)
    while lines and (not lines[-1] or lines[-1].isspace()):
        lines.pop()
    for i, line in enumerate(lines):
        lines[i] = line.rstrip()
    string = "\n".join(lines)
    string = textwrap.dedent(string)
    if indent:
        if isinstance(indent, str):
            indent_string = indent
        else:
            assert isinstance(indent, int)
            indent_string = abs(int(indent)) * " "
        lines = string.split("\n")
        for i, line in enumerate(lines):
            if line:
                lines[i] = f"{indent_string}{line}"
        string = "\n".join(lines)
    return string


def pluralize(string: str, count: int | None = None) -> str:
    """
    Pluralizes English string.

    Changes terminal ``-y`` to ``-ies``:

    ..  container:: example

        >>> abjad.string.pluralize('catenary')
        'catenaries'

        Adds ``-es`` to terminal ``-s``, ``-sh``, ``-x`` and ``-z``:

        >>> abjad.string.pluralize('brush')
        'brushes'

        Adds ``-s`` to all other strings:

        >>> abjad.string.pluralize('shape')
        'shapes'

    ..  container:: example

        Does not pluralize when ``count`` is 1:

        >>> abjad.string.pluralize('shape', count=1)
        'shape'

    """
    if count == 1:
        return string
    elif string.endswith("y"):
        result = string[:-1] + "ies"
    elif string.endswith(("s", "sh", "x", "z")):
        result = string + "es"
    else:
        result = string + "s"
    return result


def sort_roman(strings: list[str]) -> list[str]:
    """
    Sorts strings containing Roman numerals.

    ..  container:: example

        >>> strings = ['TromboneII', 'TromboneIII', 'TromboneI']
        >>> abjad.string.sort_roman(strings)
        ['TromboneI', 'TromboneII', 'TromboneIII']

        >>> strings = ['ViolinXI', 'ViolinX', 'ViolinIX']
        >>> abjad.string.sort_roman(strings)
        ['ViolinIX', 'ViolinX', 'ViolinXI']

    """
    lists = []
    for string in strings:
        list_ = []
        for word in delimit_words(string):
            try:
                number = roman.fromRoman(word)
                list_.append(number)
            except roman.InvalidRomanNumeralError:
                list_.append(word)
        lists.append(list_)
    pairs = list(zip(strings, lists))
    pairs.sort(key=lambda pair: pair[1])
    strings_ = [pair[0] for pair in pairs]
    return strings_


def strip_roman(string: str) -> str:
    """
    Strips roman numerals from right of string.

    ..  container:: example

        >>> abjad.string.strip_roman('Trombone')
        'Trombone'

        >>> abjad.string.strip_roman('TromboneI')
        'Trombone'

        >>> abjad.string.strip_roman('TromboneII')
        'Trombone'

        >>> abjad.string.strip_roman('TromboneIII')
        'Trombone'

        >>> abjad.string.strip_roman('TromboneIV')
        'Trombone'

    """
    words = delimit_words(string)
    try:
        roman.fromRoman(words[-1])
        words = words[:-1]
    except roman.InvalidRomanNumeralError:
        pass
    return "".join(words)


def to_shout_case(string: str) -> str:
    """
    Changes string to shout case.

    ..  container:: example

        >>> abjad.string.to_shout_case('scale degrees 4 and 5')
        'SCALE_DEGREES_4_AND_5'

        >>> abjad.string.to_shout_case('scale_degrees_4_and_5')
        'SCALE_DEGREES_4_AND_5'

        >>> abjad.string.to_shout_case('scale-degrees-4-and-5')
        'SCALE_DEGREES_4_AND_5'

        >>> abjad.string.to_shout_case('ScaleDegrees4And5')
        'SCALE_DEGREES_4_AND_5'

    """
    words = delimit_words(string)
    words_ = [_.upper() for _ in words]
    string = "_".join(words_)
    return string


def to_tridirectional_lilypond_symbol(
    argument: int | str | None,
) -> str | None:
    """
    Changes ``argument`` to tridirectional LilyPond symbol.

    ..  container:: example

        >>> abjad.string.to_tridirectional_lilypond_symbol(abjad.UP)
        '^'

        >>> abjad.string.to_tridirectional_lilypond_symbol(abjad.DOWN)
        '_'

        >>> abjad.string.to_tridirectional_lilypond_symbol(1)
        '^'

        >>> abjad.string.to_tridirectional_lilypond_symbol(0)
        '-'

        >>> abjad.string.to_tridirectional_lilypond_symbol(-1)
        '_'

        >>> abjad.string.to_tridirectional_lilypond_symbol('^')
        '^'

        >>> abjad.string.to_tridirectional_lilypond_symbol('-')
        '-'

        >>> abjad.string.to_tridirectional_lilypond_symbol('_')
        '_'

        Returns none when ``argument`` is none:

        >>> abjad.string.to_tridirectional_lilypond_symbol(None) is None
        True

    """
    if argument is None:
        return None
    elif argument in (_enums.UP, 1, "^"):
        return "^"
    elif argument in (_enums.DOWN, -1, "_"):
        return "_"
    elif argument in (_enums.CENTER, 0, "-"):
        return "-"
    else:
        raise ValueError(repr(argument))


def to_tridirectional_ordinal_constant(
    argument: int | str | None,
) -> _enums.Vertical | None:
    """
    Changes ``argument`` to tridirectional ordinal constant.

    ..  container:: example

        >>> abjad.string.to_tridirectional_ordinal_constant('^')
        <Vertical.UP: 1>

        >>> abjad.string.to_tridirectional_ordinal_constant('_')
        <Vertical.DOWN: -1>

        >>> abjad.string.to_tridirectional_ordinal_constant(1)
        <Vertical.UP: 1>

        >>> abjad.string.to_tridirectional_ordinal_constant(-1)
        <Vertical.DOWN: -1>

        >>> abjad.string.to_tridirectional_ordinal_constant(abjad.UP)
        <Vertical.UP: 1>

        >>> abjad.string.to_tridirectional_ordinal_constant(abjad.DOWN)
        <Vertical.DOWN: -1>

        >>> abjad.string.to_tridirectional_ordinal_constant(abjad.CENTER)
        <Vertical.CENTER: 0>

        >>> abjad.string.to_tridirectional_ordinal_constant(None) is None
        True

    """
    if argument is None:
        return None
    elif argument in ("^", _enums.UP, 1):
        return _enums.UP
    elif argument in ("_", _enums.DOWN, -1):
        return _enums.DOWN
    elif argument in ("-", _enums.CENTER, 0):
        return _enums.CENTER
    else:
        raise ValueError(repr(argument))


def to_upper_camel_case(string: str) -> str:
    """
    Changes string to upper camel case.

    ..  container:: example

        >>> abjad.string.to_upper_camel_case('scale degrees 4 and 5')
        'ScaleDegrees4And5'

        >>> abjad.string.to_upper_camel_case('scale_degrees_4_and_5')
        'ScaleDegrees4And5'

        >>> abjad.string.to_upper_camel_case('scale-degrees-4-and-5')
        'ScaleDegrees4And5'

        >>> abjad.string.to_upper_camel_case('ScaleDegrees4And5')
        'ScaleDegrees4And5'

    """
    words = delimit_words(string)
    words_ = [_.capitalize() for _ in words]
    string = "".join(words_)
    return string



import pytest

import abjad


def test_indicators_01():
    """
    Classes ignore "hide" in definitions of __eq__.

    Test will be removed when "hide" migrates from indicators to wrapper.
    """

    assert abjad.Clef("treble", hide=False) == abjad.Clef("treble", hide=False)
    assert abjad.Clef("treble", hide=False) == abjad.Clef("treble", hide=True)
    assert abjad.Clef("treble", hide=True) == abjad.Clef("treble", hide=False)
    assert abjad.Clef("treble", hide=True) == abjad.Clef("treble", hide=True)
    assert abjad.Clef("treble", hide=False) != abjad.Clef("alto", hide=False)
    assert abjad.Clef("treble", hide=False) != abjad.Clef("alto", hide=True)
    assert abjad.Clef("treble", hide=True) != abjad.Clef("alto", hide=False)
    assert abjad.Clef("treble", hide=True) != abjad.Clef("alto", hide=True)

    assert abjad.Dynamic("p", hide=False) == abjad.Dynamic("p", hide=False)
    assert abjad.Dynamic("p", hide=False) == abjad.Dynamic("p", hide=True)
    assert abjad.Dynamic("p", hide=True) == abjad.Dynamic("p", hide=False)
    assert abjad.Dynamic("p", hide=True) == abjad.Dynamic("p", hide=True)
    assert abjad.Dynamic("p", hide=False) != abjad.Dynamic("f", hide=False)
    assert abjad.Dynamic("p", hide=False) != abjad.Dynamic("f", hide=True)
    assert abjad.Dynamic("p", hide=True) != abjad.Dynamic("f", hide=False)
    assert abjad.Dynamic("p", hide=True) != abjad.Dynamic("f", hide=True)

    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=False
    ) == abjad.MetronomeMark(abjad.Duration(1, 4), 60, hide=False)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=False
    ) == abjad.MetronomeMark(abjad.Duration(1, 4), 60, hide=True)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=True
    ) == abjad.MetronomeMark(abjad.Duration(1, 4), 60, hide=False)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=True
    ) == abjad.MetronomeMark(abjad.Duration(1, 4), 60, hide=True)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=False
    ) != abjad.MetronomeMark(abjad.Duration(1, 4), 72, hide=False)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=False
    ) != abjad.MetronomeMark(abjad.Duration(1, 4), 72, hide=True)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=True
    ) != abjad.MetronomeMark(abjad.Duration(1, 4), 72, hide=False)
    assert abjad.MetronomeMark(
        abjad.Duration(1, 4), 60, hide=True
    ) != abjad.MetronomeMark(abjad.Duration(1, 4), 72, hide=True)

    assert abjad.TimeSignature(((3, 4)), hide=False) == abjad.TimeSignature(
        ((3, 4)), hide=False
    )
    assert abjad.TimeSignature(((3, 4)), hide=False) == abjad.TimeSignature(
        ((3, 4)), hide=True
    )
    assert abjad.TimeSignature(((3, 4)), hide=True) == abjad.TimeSignature(
        ((3, 4)), hide=False
    )
    assert abjad.TimeSignature(((3, 4)), hide=True) == abjad.TimeSignature(
        ((3, 4)), hide=True
    )
    assert abjad.TimeSignature(((3, 4)), hide=False) != abjad.TimeSignature(
        ((4, 4)), hide=False
    )
    assert abjad.TimeSignature(((3, 4)), hide=False) != abjad.TimeSignature(
        ((4, 4)), hide=True
    )
    assert abjad.TimeSignature(((3, 4)), hide=True) != abjad.TimeSignature(
        ((4, 4)), hide=False
    )
    assert abjad.TimeSignature(((3, 4)), hide=True) != abjad.TimeSignature(
        ((4, 4)), hide=True
    )


# TODO: parameterize each test below:


def test_indicators_02():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.hide == indicator_2.hide == False
    Result:
        * raises exception
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble"), staff[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.Clef("treble"), staff[0])
    assert "attempting to attach conflicting indicator" in str(e)
    """
    abjad.attach(abjad.VoiceNumber(1), staff[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.VoiceNumber(1), staff[0])
    assert "attempting to attach conflicting indicator" in str(e)
    """


def test_indicators_03():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 != indicator_2
        * indicator_1.hide == indicator_2.hide == False
    Result:
        * raises exception
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble"), staff[0])
    with pytest.raises(Exception):
        abjad.attach(abjad.Clef("alto"), staff[0])


def test_indicators_04():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.hide == indicator_2.hide == True
    Result:
        * raises exception
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble", hide=True), staff[0])
    with pytest.raises(Exception):
        abjad.attach(abjad.Clef("treble", hide=True), staff[0])


def test_indicators_05():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.hide == False and indicator_2.hide == True
    Result:
        * raises exception
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble"), staff[0])
    with pytest.raises(Exception):
        abjad.attach(abjad.Clef("treble", hide=True), staff[0])


def test_indicators_06():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.hide == True and indicator_2.hide == False
    Result:
        * raises exception
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble", hide=True), staff[0])
    with pytest.raises(Exception):
        abjad.attach(abjad.Clef("treble"), staff[0])


def test_indicators_07():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 =! indicator_2
        * indicator_1.hide == False and indicator_2.hide == True
    Result:
        * ok
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble"), staff[0])
    abjad.attach(abjad.Clef("alto", hide=True), staff[0])


def test_indicators_08():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 != indicator_2
        * indicator_1.hide == True and indicator_2.hide == False
    Result:
        * ok
    """
    staff = abjad.Staff("c'4 d' e' f'")
    abjad.attach(abjad.Clef("treble", hide=True), staff[0])
    abjad.attach(abjad.Clef("alto"), staff[0])


def test_indicators_09():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.site != indicator_2.site
    Result:
        * ok
    """
    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.BarLine("|", site="before"), voice[0])
    abjad.attach(abjad.BarLine("|", site="after"), voice[0])

    # TODO: this should raise some other type of exception
    #       having to do with redundancy because ottava 1
    #       followed later by ottava 1 is redundant;
    #       this arises only for persistent indicators
    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.Ottava(1, site="before"), voice[0])
    abjad.attach(abjad.Ottava(1, site="after"), voice[0])


def test_indicators_10():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 != indicator_2
        * indicator_1.site != indicator_2.site
    Result:
        * ok
    """
    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.BarLine("|", site="before"), voice[0])
    abjad.attach(abjad.BarLine("||", site="after"), voice[0])

    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.Ottava(1, site="before"), voice[0])
    abjad.attach(abjad.Ottava(0, site="after"), voice[0])


def test_indicators_11():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 == indicator_2
        * indicator_1.site == indicator_2.site
    Result:
        * exception
    """
    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.BarLine("|", site="after"), voice[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.BarLine("|", site="after"), voice[0])
    assert "attempting to attach conflicting indicator" in str(e)

    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.Ottava(1, site="before"), voice[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.Ottava(1, site="before"), voice[0])
    assert "attempting to attach conflicting indicator" in str(e)


def test_indicators_12():
    """
    Conditions:
        * indicator_1, indicator_2 attach to same leaf
        * class(indicator_1) is class(indicator_2)
        * indicator_1 != indicator_2
        * indicator_1.site == indicator_2.site
    Result:
        * exception
    """
    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.BarLine("|", site="after"), voice[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.BarLine("||", site="after"), voice[0])
    assert "attempting to attach conflicting indicator" in str(e)

    voice = abjad.Voice("c'4", name="Voice")
    staff = abjad.Staff([voice], name="Staff")
    abjad.Score([staff], name="Score")
    abjad.attach(abjad.Ottava(1, site="before"), voice[0])
    with pytest.raises(Exception) as e:
        abjad.attach(abjad.Ottava(0, site="before"), voice[0])
    assert "attempting to attach conflicting indicator" in str(e)


def test_allowable_sites():
    """
    Indicators with "allowable_sites" property behave like this.
    """

    voice = abjad.Voice("c'4")
    literal = abjad.LilyPondLiteral("% leaf comment", site="before")
    abjad.attach(literal, voice[0])
    literal = abjad.LilyPondLiteral("% container comment", site="opening")
    abjad.attach(literal, voice)
    literal = abjad.LilyPondLiteral("% container comment", site="opening")
    with pytest.raises(Exception) as e:
        abjad.attach(literal, voice[0])
    assert "allows only" in str(e)


def test_StartTrillSpan_01():
    """
    Set force_trill_pitch_head_accidental=True to force trill pitch head accidental.
    """

    voice = abjad.Voice("c'4 d' e' f'")
    start_trill_span = abjad.StartTrillSpan(
        pitch=abjad.NamedPitch("D4"),
        force_trill_pitch_head_accidental=True,
    )
    abjad.attach(start_trill_span, voice[0])
    stop_trill_span = abjad.StopTrillSpan()
    abjad.attach(stop_trill_span, voice[-1])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \pitchedTrill
            c'4
            \startTrillSpan d'!
            d'4
            e'4
            f'4
            \stopTrillSpan
        }
        """
    )



import abjad


def test_Cluster___init___01():
    """
    Cluster can be empty.
    """
    cluster = abjad.Cluster([])
    assert not cluster.simultaneous
    assert len(cluster) == 0
    assert abjad.lilypond(cluster) == abjad.string.normalize(
        r"""
        \makeClusters {
        }
        """
    )


def test_Cluster___init___02():
    cluster = abjad.Cluster("cs'4 cs'4 cs'4 cs'4")
    assert isinstance(cluster, abjad.Cluster)
    assert not cluster.simultaneous
    assert len(cluster) == 4
    assert abjad.lilypond(cluster) == abjad.string.normalize(
        r"""
        \makeClusters {
            cs'4
            cs'4
            cs'4
            cs'4
        }
        """
    )



import abjad


def test_mutate_replace_01():
    """
    Moves parentage from two old notes to five new notes.

    Equivalent to staff[1:3] = new_notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])
    abjad.beam(voice[2:])
    abjad.hairpin("< !", voice[:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \<
            d'8
            ]
            e'8
            [
            f'8
            \!
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    old_notes = voice[1:3]
    new_notes = [
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
    ]

    abjad.mutate.replace(old_notes, new_notes)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \<
            c''16
            c''16
            c''16
            c''16
            c''16
            f'8
            \!
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_replace_02():
    """
    Moves parentage from one old note to five new notes.

    Equivalent to voice[:1] = new_notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])
    abjad.beam(voice[2:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    old_notes = voice[:1]
    new_notes = [
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
    ]
    abjad.mutate.replace(old_notes, new_notes)
    abjad.attach(abjad.StartBeam(), voice[0])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c''16
            [
            c''16
            c''16
            c''16
            c''16
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_replace_03():
    """
    Moves parentage from two old notes to five new notes.

    Equivalent to voice[:2] = new_notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])
    abjad.beam(voice[2:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    old_notes = voice[:2]
    new_notes = [
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
    ]
    abjad.mutate.replace(old_notes, new_notes)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c''16
            c''16
            c''16
            c''16
            c''16
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_replace_04():
    """
    Moves parentage from three old notes to five new notes.

    Equivalent to voice[:3] = new_notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])
    abjad.beam(voice[2:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    old_notes = voice[:3]
    new_notes = [
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
    ]
    abjad.mutate.replace(old_notes, new_notes)
    abjad.attach(abjad.StartBeam(), voice[0])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c''16
            [
            c''16
            c''16
            c''16
            c''16
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_replace_05():
    """
    Moves parentage from four old notes to five new notes.

    Equivalent to voice[:] = new_notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])
    abjad.beam(voice[2:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    old_notes = voice[:]
    new_notes = [
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
        abjad.Note("c''16"),
    ]
    abjad.mutate.replace(old_notes, new_notes)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c''16
            c''16
            c''16
            c''16
            c''16
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_replace_06():
    """
    Moves parentage from container to children of container.

    Replaces container with contents of container.

    Effectively removes container from score.

    Equivalent to staff[:1] = staff[0][:].
    """

    staff = abjad.Staff([abjad.Voice("c'8 d'8 e'8 f'8")])
    abjad.beam(staff[0][:])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                [
                d'8
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    voice_selection = staff[:1]
    voice = voice_selection[0]
    abjad.mutate.replace(voice_selection, staff[0][:])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(staff))

    assert not voice
    assert abjad.wf.wellformed(staff)



import copy
import pickle

import pytest

import abjad


def test_RhythmTreeContainer___call___01():
    rtm = "(1 (1 (2 (1 1 1)) 2))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
    result = tree((1, 4))
    assert isinstance(result, list)
    assert len(result) == 1
    assert abjad.lilypond(result[0]) == abjad.string.normalize(
        r"""
        \tuplet 5/4
        {
            c'16
            \tuplet 3/2
            {
                c'16
                c'16
                c'16
            }
            c'8
        }
        """
    )


def test_RhythmTreeContainer___call___02():
    rtm = "(1 (1 (2 (1 1 1 1)) 1))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
    components = tree((1, 4))
    tuplet = components[0]._parent
    staff = abjad.Staff([tuplet])
    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                c'16
                c'32
                c'32
                c'32
                c'32
                c'16
            }
        }
        """
    ), print(staff)


def test_RhythmTreeContainer___contains___01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1, 1))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1, 1), children=[leaf_b]
    )
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1, 1), children=[leaf_a, subcontainer]
    )
    assert leaf_a in container
    assert leaf_b not in container
    assert leaf_c not in container


def test_RhythmTreeContainer___eq___01():
    a = abjad.rhythmtrees.RhythmTreeContainer(children=[])
    b = abjad.rhythmtrees.RhythmTreeContainer(children=[])
    assert repr(a) == repr(b)
    assert a != b


def test_RhythmTreeContainer___eq___02():
    a = abjad.rhythmtrees.RhythmTreeContainer(
        children=[abjad.rhythmtrees.RhythmTreeLeaf()]
    )
    b = abjad.rhythmtrees.RhythmTreeContainer(
        children=[abjad.rhythmtrees.RhythmTreeLeaf()]
    )
    assert repr(a) == repr(b)
    assert a != b


def test_RhythmTreeContainer___eq___03():
    a = abjad.rhythmtrees.RhythmTreeContainer(children=[])
    b = abjad.rhythmtrees.RhythmTreeContainer(preprolated_duration=(2, 1), children=[])
    c = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1),
        children=[abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))],
    )
    d = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(2, 1),
        children=[abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))],
    )
    e = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(2, 1),
        children=[abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))],
    )
    assert a != b
    assert a != c
    assert a != d
    assert a != e
    assert b != c
    assert b != d
    assert b != e
    assert c != d
    assert c != e
    assert d != e


def test_RhythmTreeContainer___getitem___01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, leaf_b, leaf_c]
    )
    assert container[0] is leaf_a
    assert container[1] is leaf_b
    assert container[2] is leaf_c
    with pytest.raises(Exception):
        container[3]
    assert container[-1] is leaf_c
    assert container[-2] is leaf_b
    assert container[-3] is leaf_a
    with pytest.raises(Exception):
        container[-4]


def test_RhythmTreeContainer___init___01():
    container = abjad.rhythmtrees.RhythmTreeContainer()
    assert container.children == ()
    assert container.preprolated_duration == 1
    assert container.start_offset == 0
    assert container.parent is None


def test_RhythmTreeContainer___init___02():
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(2), children=[]
    )
    assert container.children == ()
    assert container.preprolated_duration == 2
    assert container.start_offset == 0
    assert container.parent is None


def test_RhythmTreeContainer___init___03():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(2))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1))
    assert leaf_a.start_offset == 0
    assert leaf_a.parent is None
    assert leaf_b.start_offset == 0
    assert leaf_b.parent is None
    assert leaf_c.start_offset == 0
    assert leaf_c.parent is None
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(4), children=[leaf_a, leaf_b, leaf_c]
    )
    assert container.children == (leaf_a, leaf_b, leaf_c)
    assert container.preprolated_duration == 4
    assert container.start_offset == 0
    assert container.parent is None
    assert leaf_a.start_offset == 0
    assert leaf_a.parent is container
    assert leaf_b.start_offset == 1
    assert leaf_b.parent is container
    assert leaf_c.start_offset == 3
    assert leaf_c.parent is container


def test_RhythmTreeContainer___iter___01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, leaf_b, leaf_c]
    )
    assert [_ for _ in container] == [leaf_a, leaf_b, leaf_c]


def test_RhythmTreeContainer___len___01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(2, 1), children=[leaf_b, leaf_c]
    )
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, subcontainer, leaf_d]
    )
    assert len(container) == 3


def test_RhythmTreeContainer_append_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer()
    assert container.children == ()
    container.append(leaf_a)
    assert container.children == (leaf_a,)
    container.append(leaf_b)
    assert container.children == (leaf_a, leaf_b)
    container.append(leaf_c)
    assert container.children == (leaf_a, leaf_b, leaf_c)
    container.append(leaf_d)
    assert container.children == (leaf_a, leaf_b, leaf_c, leaf_d)
    container.append(leaf_a)
    assert container.children == (leaf_b, leaf_c, leaf_d, leaf_a)


def test_RhythmTreeContainer_children_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(2, 1), children=[leaf_b, leaf_c]
    )
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, subcontainer, leaf_d]
    )
    assert container.children == (leaf_a, subcontainer, leaf_d)


def test_RhythmTreeContainer_contents_duration_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(3))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(3))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(2))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(2), children=[leaf_b, leaf_c]
    )
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1), children=[leaf_a, subcontainer, leaf_d]
    )
    assert container._get_contents_duration() == 6
    assert subcontainer._get_contents_duration() == 5


def test_RhythmTreeContainer_extend_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer()
    assert container.children == ()
    container.extend([leaf_a])
    assert container.children == (leaf_a,)
    container.extend([leaf_b, leaf_c, leaf_d])
    assert container.children == (leaf_a, leaf_b, leaf_c, leaf_d)
    container.extend([leaf_a, leaf_c])
    assert container.children == (leaf_b, leaf_d, leaf_a, leaf_c)


def test_RhythmTreeContainer_index_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(2, 1), children=[leaf_b, leaf_c]
    )
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, subcontainer, leaf_d]
    )
    assert container.index(leaf_a) == 0
    assert container.index(subcontainer) == 1
    assert container.index(leaf_d) == 2
    with pytest.raises(ValueError):
        container.index(leaf_b)
    with pytest.raises(ValueError):
        container.index(leaf_c)


def test_RhythmTreeContainer_insert_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer()
    assert container.children == ()
    container.insert(0, leaf_a)
    assert container.children == (leaf_a,)
    container.insert(0, leaf_b)
    assert container.children == (leaf_b, leaf_a)
    container.insert(1, leaf_c)
    assert container.children == (leaf_b, leaf_c, leaf_a)


def test_RhythmTreeContainer_pop_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, leaf_b, leaf_c]
    )
    assert container.children == (leaf_a, leaf_b, leaf_c)
    assert leaf_a.parent is container
    assert leaf_b.parent is container
    assert leaf_c.parent is container
    result = container.pop()
    assert container.children == (leaf_a, leaf_b)
    assert result is leaf_c
    assert result.parent is None
    result = container.pop(0)
    assert container.children == (leaf_b,)
    assert result is leaf_a
    assert result.parent is None


def test_RhythmTreeContainer_remove_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=(1, 1), children=[leaf_a, leaf_b, leaf_c]
    )
    assert container.children == (leaf_a, leaf_b, leaf_c)
    assert leaf_a.parent is container
    assert leaf_b.parent is container
    assert leaf_c.parent is container
    container.remove(leaf_a)
    assert container.children == (leaf_b, leaf_c)
    assert leaf_a.parent is None
    container.remove(leaf_c)
    assert container.children == (leaf_b,)
    assert leaf_c.parent is None


def test_RhythmTreeContainer_rtm_format_01():
    leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(3, 1))
    leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(3, 1))
    leaf_c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(2, 1))
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(2, 1), children=[leaf_b, leaf_c]
    )
    leaf_d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1, 1))
    container = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1, 1),
        children=[leaf_a, subcontainer, leaf_d],
    )
    assert subcontainer.rtm_format == "(2 (3 2))"
    assert container.rtm_format == "(1 (3 (2 (3 2)) 1))"


def test_RhythmTreeLeaf___copy___01():
    leaf = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1))
    copied = copy.copy(leaf)
    assert repr(leaf) == repr(copied)
    assert leaf is not copied


def test_RhythmTreeLeaf___copy___02():
    leaf = abjad.rhythmtrees.RhythmTreeLeaf(
        preprolated_duration=(2, 1), is_pitched=True
    )
    copied = copy.copy(leaf)
    assert repr(leaf) == repr(copied)
    assert leaf is not copied


def test_RhythmTreeLeaf___eq___01():
    a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1), is_pitched=True)
    b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1), is_pitched=True)
    assert repr(a) == repr(b)
    assert a != b


def test_RhythmTreeLeaf___eq___02():
    a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1), is_pitched=True)
    b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(1, 1), is_pitched=False)
    c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1), is_pitched=True)
    d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=(2, 1), is_pitched=False)
    assert a != b
    assert a != c
    assert a != d
    assert b != c
    assert b != d
    assert c != d


def test_RhythmTreeNode___call___01():
    rtm = "(1 (1 1 1 1))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
    components = tree((1, 4))
    assert len(components) == 4
    assert all(isinstance(_, abjad.Note) for _ in components)
    assert all(_.written_duration == abjad.Duration(1, 16) for _ in components)


def test_RhythmTreeNode___call___02():
    rtm = "(1 (1 (2 (1 1 1)) 2))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
    result = tree((1, 4))
    assert isinstance(result, list)
    assert len(result) == 1
    assert abjad.lilypond(result[0]) == abjad.string.normalize(
        r"""
        \tuplet 5/4
        {
            c'16
            \tuplet 3/2
            {
                c'16
                c'16
                c'16
            }
            c'8
        }
        """
    )


def test_RhythmTreeNode___call___03():
    rtm = "(1 (1 (2 (1 (2 (1 1)) 1)) 2))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
    result = tree((1, 4))
    assert abjad.lilypond(result[0]) == abjad.string.normalize(
        r"""
        \tuplet 5/4
        {
            c'16
            c'32
            c'32
            c'32
            c'32
            c'8
        }
        """
    )


def test_RhythmTreeNode_depth_01():
    container = abjad.rhythmtrees.RhythmTreeContainer()
    assert container.depth == 0
    leaf = abjad.rhythmtrees.RhythmTreeLeaf()
    assert leaf.depth == 0
    container.append(leaf)
    assert leaf.depth == 1
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer()
    assert subcontainer.depth == 0
    container.append(subcontainer)
    assert subcontainer.depth == 1
    subcontainer.append(leaf)
    assert leaf.depth == 2
    subsubcontainer = abjad.rhythmtrees.RhythmTreeContainer()
    assert subsubcontainer.depth == 0
    subcontainer.append(subsubcontainer)
    assert subsubcontainer.depth == 2
    subsubcontainer.append(leaf)
    assert leaf.depth == 3


def test_RhythmTreeNode_duration_01():
    tree = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1),
        children=[
            abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1)),
            abjad.rhythmtrees.RhythmTreeContainer(
                preprolated_duration=abjad.Duration(2),
                children=[
                    abjad.rhythmtrees.RhythmTreeLeaf(
                        preprolated_duration=abjad.Duration(3)
                    ),
                    abjad.rhythmtrees.RhythmTreeLeaf(
                        preprolated_duration=abjad.Duration(2)
                    ),
                ],
            ),
            abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(2)),
        ],
    )
    assert tree.duration == abjad.Duration(1)
    assert tree[0].duration == abjad.Duration(1, 5)
    assert tree[1].duration == abjad.Duration(2, 5)
    assert tree[1][0].duration == abjad.Duration(6, 25)
    assert tree[1][1].duration == abjad.Duration(4, 25)
    assert tree[2].duration == abjad.Duration(2, 5)
    tree[1].append(tree.pop())
    assert tree.duration == abjad.Duration(1)
    assert tree[0].duration == abjad.Duration(1, 3)
    assert tree[1].duration == abjad.Duration(2, 3)
    assert tree[1][0].duration == abjad.Duration(2, 7)
    assert tree[1][1].duration == abjad.Duration(4, 21)
    assert tree[1][2].duration == abjad.Duration(4, 21)
    tree.preprolated_duration = abjad.Duration(19)
    assert tree.duration == abjad.Duration(19)
    assert tree[0].duration == abjad.Duration(19, 3)
    assert tree[1].duration == abjad.Duration(38, 3)
    assert tree[1][0].duration == abjad.Duration(38, 7)
    assert tree[1][1].duration == abjad.Duration(76, 21)
    assert tree[1][2].duration == abjad.Duration(76, 21)


def test_RhythmTreeNode_offset_01():
    tree = abjad.rhythmtrees.RhythmTreeContainer(
        preprolated_duration=abjad.Duration(1),
        children=[
            abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(1)),
            abjad.rhythmtrees.RhythmTreeContainer(
                preprolated_duration=abjad.Duration(2),
                children=[
                    abjad.rhythmtrees.RhythmTreeLeaf(
                        preprolated_duration=abjad.Duration(3)
                    ),
                    abjad.rhythmtrees.RhythmTreeLeaf(
                        preprolated_duration=abjad.Duration(2)
                    ),
                ],
            ),
            abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=abjad.Duration(2)),
        ],
    )
    assert tree.start_offset == abjad.Offset(0)
    assert tree[0].start_offset == abjad.Offset(0)
    assert tree[1].start_offset == abjad.Offset(1, 5)
    assert tree[1][0].start_offset == abjad.Offset(1, 5)
    assert tree[1][1].start_offset == abjad.Offset(11, 25)
    assert tree[2].start_offset == abjad.Offset(3, 5)
    node = tree.pop()
    assert node.start_offset == abjad.Offset(0)
    tree[1].append(node)
    assert tree.start_offset == abjad.Offset(0)
    assert tree[0].start_offset == abjad.Offset(0)
    assert tree[1].start_offset == abjad.Offset(1, 3)
    assert tree[1][0].start_offset == abjad.Offset(1, 3)
    assert tree[1][1].start_offset == abjad.Offset(13, 21)
    assert tree[1][2].start_offset == abjad.Offset(17, 21)
    tree.preprolated_duration = abjad.Duration(19)
    assert tree.start_offset == abjad.Offset(0)
    assert tree[0].start_offset == abjad.Offset(0)
    assert tree[1].start_offset == abjad.Offset(19, 3)
    assert tree[1][0].start_offset == abjad.Offset(19, 3)
    assert tree[1][1].start_offset == abjad.Offset(247, 21)
    assert tree[1][2].start_offset == abjad.Offset(323, 21)


def test_RhythmTreeNode_parent_01():
    leaf = abjad.rhythmtrees.RhythmTreeLeaf()
    container = abjad.rhythmtrees.RhythmTreeContainer()
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer()
    assert leaf.parent is None
    assert container.parent is None
    assert subcontainer.parent is None
    container.append(leaf)
    assert leaf.parent is container
    container.append(subcontainer)
    assert subcontainer.parent is container
    assert leaf.parent is container
    assert container.parent is None
    subcontainer.append(leaf)
    assert leaf.parent is subcontainer
    assert subcontainer.parent is container
    assert container.parent is None
    with pytest.raises(ValueError):
        subcontainer.append(container)


def test_RhythmTreeNode_parentage_ratios_01():
    string = "(1 (1 (2 (3 4)) 2))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(string)[0]
    assert tree.parentage_ratios == (1,)
    assert tree[0].parentage_ratios == (1, (1, 5))
    assert tree[1].parentage_ratios == (1, (2, 5))
    assert tree[1][0].parentage_ratios == (1, (2, 5), (3, 7))
    assert tree[1][1].parentage_ratios == (1, (2, 5), (4, 7))
    assert tree[2].parentage_ratios == (1, (2, 5))


def test_RhythmTreeNode_pickle_01():
    string = "(1 (1 (2 (1 1 1)) 2))"
    tree = abjad.rhythmtrees.RhythmTreeParser()(string)[0]
    pickled = pickle.loads(pickle.dumps(tree))
    assert repr(pickled) == repr(tree)
    assert pickled != tree
    assert pickled is not tree


def test_RhythmTreeNode_root_01():
    leaf = abjad.rhythmtrees.RhythmTreeLeaf()
    subsubcontainer = abjad.rhythmtrees.RhythmTreeContainer()
    subcontainer = abjad.rhythmtrees.RhythmTreeContainer()
    container = abjad.rhythmtrees.RhythmTreeContainer()
    container.append(subcontainer)
    subcontainer.append(subsubcontainer)
    subsubcontainer.append(leaf)
    assert leaf.root is container
    assert subsubcontainer.root is container
    assert subcontainer.root is container
    assert container.root is None



import pytest

import abjad


def test_bind_01():
    """
    REGRESSION. Make sure no exceptions are raised when updating
    wrappers that have been orphaned because a context was removed.
    In this example, the metronome mark finds the score context at
    attachment time. After the score is removed, the metronome mark
    must properly update its effective context to none.
    """

    voice = abjad.Voice("c'4")
    staff = abjad.Staff([voice])
    score = abjad.Score([staff])
    leaf = abjad.select.leaf(staff, 0)
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 72)
    abjad.attach(mark, leaf)
    score[:] = []
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                \tempo 4=72
                c'4
            }
        }
        """
    )


def test_bind_02():
    """
    Make sure exception is raised when check_duplicate_indicator=True.
    """

    score = abjad.Score(r"\new Staff { c'' d'' e'' f'' } \new Staff { c' d' e' f' }")
    mark_1 = abjad.MetronomeMark(abjad.Duration(1, 8), 52)
    mark_2 = abjad.MetronomeMark(abjad.Duration(1, 8), 73)
    abjad.attach(mark_1, score[0][0])

    with pytest.raises(Exception):
        abjad.attach(mark_2, score[1][0], check_duplicate_indicator=True)
